from argparse import ArgumentParser, REMAINDER, Namespace


def add_launch_arguments(parser: ArgumentParser, _async: bool = False):
    parser.add_argument(
        "--gpus",
        "-n",
        default="all",
        type=str,
        required=_async,
        help="Number or GPU indices to use (e.g. -n=0,1,4,5 | -n=all | -n=available)."
    )
    if not _async:
        parser.add_argument(
            "-N",
            default="0",
            type=str,
            required=False,
            help="Number of GPUs to use. This does not consider GPU indices by default, although you can represent "
                    "a Python slice. (e.g. '2:', which means from index 2 to the last GPU index, or "
                    "'3:8', which means from index 3 to index 7, or lastly ':4', which means indices 0 to 3 or a total of 4 gpus)."
        )
    parser.add_argument(
        "--strat",
        type=str,
        required=False,
        default="ddp",
        help="Parallelism strategy to apply or config file path. See 'accmt strats'."
    )
    parser.add_argument("-O1", action="store_true", help="Apply optimization type 1: efficient OMP_NUM_THREADS.")
    parser.add_argument("--cpu", action="store_true", help="Destinate this process to CPU.")

    # TODO: For now, we need to find a way to collect processes that are running on certain GPUs to verify if they're free to use.
    #parser.add_argument("--ignore-warnings", action="store_true", help="Ignore warnings (launch independent if GPUs are being used).")

    parser.add_argument("--suppress-queue", "-sq", "-SQ", "-sQ", "-Sq", action="store_true")
    parser.add_argument(
        "--queue-file",
        "-Q",
        "-q",
        type=str,
        required=False,
        default=".accmt",
        help="Specify queue file (default is '.accmt')."
    )
    if _async:
        parser.add_argument(
            "--evaluation-device-indices",
            "-e",
            type=str,
            required=True,
            help="Number or GPU indices to use for evaluation (e.g. -n=0,1,4,5)."
        )
    parser.add_argument("file", type=str, help="File to run training.")
    parser.add_argument("extra_args", nargs=REMAINDER)


def get_parser() -> tuple[ArgumentParser, Namespace]:
    parser = ArgumentParser(description="AcceleratorModule CLI to run train processes on top of 🤗 Accelerate.")
    subparsers = parser.add_subparsers(dest="command", help="Available subcommands")

    # Run distributed training
    launch_parsers, debug_parsers = [], []

    launch_parsers.append(subparsers.add_parser("launch", help="Launch distributed training processes."))
    launch_parsers.append(subparsers.add_parser("alaunch", help="Launch distributed training processes with asynchronous evaluation."))
    launch_parsers.append(subparsers.add_parser("async-launch", help="Launch distributed training processes with asynchronous evaluation."))
    debug_parsers.append(subparsers.add_parser("debug", help="Launch distributed training processes in debug mode."))
    debug_parsers.append(subparsers.add_parser("adebug", help="Launch distributed training processes with asynchronous evaluation in debug mode."))
    debug_parsers.append(subparsers.add_parser("async-debug", help="Launch distributed training processes with asynchronous evaluation in debug mode."))

    for launch_parser in launch_parsers:
        _async = launch_parser.prog.split(" ")[-1] in {"alaunch", "async-launch"}
        add_launch_arguments(launch_parser, _async=_async)

    for debug_parser in debug_parsers:
        _async = debug_parser.prog.split(" ")[-1] in {"adebug", "async-debug"}
        add_launch_arguments(debug_parser, _async=_async)
        debug_parser.add_argument(
            "--level",
            "-L",
            "-l",
            type=int,
            default=4,
            required=False,
            help="Debug mode level. See more details using 'accmt debug-levels'."
        )

    # Get model from checkpoint
    get_parser = subparsers.add_parser("get", help="Get model from a checkpoint directory.")
    get_parser.add_argument("checkpoint", type=str, help="Checkpoint directory.")
    get_parser.add_argument("--out", "-O", "-o", required=True, type=str, help="Output directory path name.")
    get_parser.add_argument("--dtype", type=str, default="float32", help=(
        "Data type of model parameters. Available options are all "
        "those from PyTorch ('float32', 'float16', etc)."
    ))

    # Strats
    strats_parser = subparsers.add_parser("strats", help="Available strategies.")
    strats_parser.add_argument("--ddp", action="store_true", help="Only show DistributedDataParallel (DDP) strategies.")
    strats_parser.add_argument("--fsdp", action="store_true", help="Only show FullyShardedDataParallel (FSDP) strategies.")
    strats_parser.add_argument("--deepspeed", action="store_true", help="Only show DeepSpeed strategies.")

    # Debug levels
    debug_levels_parser = subparsers.add_parser("debug-levels", help="Available debug levels.")
    debug_levels_parser.add_argument(
        "--level",
        "-L",
        "-l",
        type=int,
        required=False,
        help="See details about a specific debug mode level."
    )

    # Generate example
    subparsers.add_parser("example", help="Generate example file.")

    return parser, parser.parse_args()



import yaml
import os
import socket
import shutil

configs = {}
_directory = os.path.dirname(__file__)
for file in os.listdir(f"{_directory}/config"):
    key = file.split(".")[0]
    configs[key] = f"{_directory}/config/{file}"

def get_free_gpus(num_devices: int) -> list[str]:
    import torch

    GB = 1024**3

    devices = []
    for i in range(num_devices):
        mem_total, mem_alloc, mem_resvd = get_cuda_device_memory(i)
        if mem_total > 0:
            device_name = torch.cuda.get_device_name(i)
            print("------------------------------------------------")
            print("GPU in use:")
            print(f"Name: {device_name}")
            print(f"Memory allocated: {mem_alloc/GB} GB")
            print(f"Memory reserved: {mem_resvd/GB} GB")
            print("ACCMT will not use this GPU during training.")
            print("------------------------------------------------")

            continue

        devices.append(str(i))

    return devices

def check_port_available(port: int, host="127.0.0.1"):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
        sock.settimeout(1)
        result = sock.connect_ex((host, port))
        return result != 0

def modify_config_file(path: str, num_gpus: int, port: int = 29500, copy: bool = False) -> tuple[int, str]:
    data = yaml.safe_load(open(path))

    if port != -1:
        _port = port
        port = port if check_port_available(port) else 0
        if port == 0:
            for current_port in range(_port+1, 65536):
                if check_port_available(current_port):
                    port = current_port
                    break
            
            if port == 0: # if 29500 to 65535 is not available
                for current_port in range(1, _port):
                    if check_port_available(current_port):
                        port = current_port
                        break

            if port == 0:
                raise RuntimeError("There are no ports available in your system.")
    else:
        port = 0  # accelerate will automatically check for an available port
    
    prev_main_process_port = data["main_process_port"] if "main_process_port" in data else -1
    prev_num_processes = data["num_processes"]

    if prev_main_process_port == port and prev_num_processes == num_gpus:
        return port, path  # skip write process

    data["main_process_port"] = port
    data["num_processes"] = num_gpus

    if copy:
        base_dir, filename = os.path.split(path)
        filename = f"_{filename}"
        path = os.path.join(base_dir, filename)

    with open(path, "w") as f:
        yaml.safe_dump(data, f)

    return port, path

def get_python_cmd():
    if shutil.which("python") is not None:
        return "python"
    else:
        return "python3"

def remove_compiled_prefix(state_dict):
    compiled = "_orig_mod" in list(state_dict.keys())[0]
    if not compiled: return state_dict

    t = type(state_dict)
    return t({k.removeprefix("_orig_mod."):v for k, v in state_dict.items()})

def show_strategies(filter: str = None):
    if filter is None: filter = ""
    for strat in configs.keys():
        if filter in strat:
            print(f"\t{strat}")

    exit(1)

def generate_hps():
    directory = os.path.dirname(__file__)
    shutil.copy(f"{directory}/example/hps_example.yaml", ".")

def get_cuda_device_memory(device: int | str) -> tuple[float, float, float]:
    import torch
    device = f"cuda:{device}" if isinstance(device, int) else device

    mem_alloc = torch.cuda.memory_allocated(device)
    mem_resvd = torch.cuda.memory_reserved(device)
    mem_total = mem_alloc + mem_resvd

    return mem_total, mem_alloc, mem_resvd

def cuda_device_in_use(device: int | str):
    mem_total, _, _ = get_cuda_device_memory(device)
    return mem_total > 0

def remove_first_line_in_file(file: str):
    with open(file, "r+") as f:
        lines = f.readlines()
        f.seek(0)
        f.writelines(lines[1:])
        f.truncate()

def get_cmd_as_list(cmd: str) -> list[str]:
    cmd = cmd.split(" ")
    cmd = [c for c in cmd if c not in {"", " "}]

    return cmd

DEBUG_LEVEL_INFO = {
    1: "Disables logging (MLFlow, Tensorboard, etc).",
    2: "Disables model and teacher compilation.",
    3: "Disables model saving, checkpointing and resuming (no folders will be created).",
    4: "Force 'eval_when_start' (in Trainer) to False.",
    5: "Disables any evaluation."
}


import hashlib
import os
import subprocess
import torch
from .utils import get_free_gpus, configs, modify_config_file, get_cmd_as_list
from datetime import datetime
import time
import signal
import sys

def launch(args):
    cpu = False
    _async = False

    current_time = datetime.now().isoformat()
    hash_object = hashlib.sha256(current_time.encode())
    hash_hex = hash_object.hexdigest()
    os.environ["ACCMT_HASH"] = hash_hex

    if "debug" in args.command:
        os.environ["ACCMT_DEBUG_MODE"] = str(args.level)

    if args.command in {"alaunch", "async-launch", "adebug", "async-debug"}:
        os.environ["ACCMT_ASYNC"] = "1"
        _async = True

    if args.cpu:
        os.environ["ACCMT_CPU"] = "1"
        cpu = True

    gpus = args.gpus.lower()
    strat = args.strat
    file = args.file
    extra_args = " ".join(args.extra_args)

    if "." in strat:
        accelerate_config_file = strat
    else:
        accelerate_config_file = configs[strat]

    if not cpu:
        if not torch.cuda.is_available():
            raise ImportError("Could not run CLI: CUDA is not available on your PyTorch installation.")
        
        NUM_DEVICES = torch.cuda.device_count()

        gpu_indices = ""
        if gpus == "available":
            gpu_indices = ",".join(get_free_gpus(NUM_DEVICES))
        elif gpus == "all":
            gpu_indices = ",".join(str(i) for i in range(NUM_DEVICES))
        else:
            gpu_indices = gpus.removeprefix(",").removesuffix(",")

        if gpu_indices == "":
            raise RuntimeError("Could not get GPU indices. If you're using 'available' in 'gpus' "
                            "parameter, make sure there is at least one GPU free of memory.")

        if not _async and args.N != "0":
            if ":" in args.N:
                _slice = slice(*map(lambda x: int(x.strip()) if x.strip() else None, args.N.split(':')))
                gpu_indices = ",".join([str(i) for i in range(NUM_DEVICES)][_slice])
            else:
                gpu_indices = ",".join(str(i) for i in range(int(args.N)))

        # TODO: For now, we need to find a way to collect processes that are running on certain GPUs to verify if they're free to use.
        #if not args.ignore_warnings:
        #    gpu_indices_list = [int(idx) for idx in gpu_indices.split(",")]
        #    device_indices_in_use = []
        #    for idx in gpu_indices_list:
        #        if cuda_device_in_use(idx):
        #            device_indices_in_use.append(idx)
        #
        #    if len(device_indices_in_use) > 0:
        #        raise RuntimeError(
        #            f"The following CUDA devices are in use: {device_indices_in_use}."
        #             "You can ignore this warning via '--ignore-warnings'."
        #        )
        
        num_processes = len(gpu_indices.split(","))
    else:
        if args.N == "0":
            raise RuntimeError("When running on CPU, '-N' must specify the number of processes to run.")
        
        if _async:
            raise NotImplementedError("Asynchronous evaluations are not supported for CPU.")

        num_processes = int(args.N)
    
    port, _ = modify_config_file(accelerate_config_file, num_processes)

    cmd = f"accelerate launch --config_file={accelerate_config_file} {file} {extra_args}"

    if not cpu:
        os.environ["CUDA_VISIBLE_DEVICES"] = gpu_indices

    if not _async:
        os.system(cmd)
    else:
        train_group = os.environ.copy()
        train_group["ACCMT_TRAIN_GROUP"] = "1"
        if args.O1:
            train_group["OMP_NUM_THREADS"] = str(os.cpu_count() // num_processes)
        cmd = get_cmd_as_list(cmd)
        try:
            process1 = subprocess.Popen(cmd, env=train_group, start_new_session=True)
            process2 = None

            if _async:
                eval_group = os.environ.copy()
                gpu_indices = args.evaluation_device_indices.removeprefix(",").removesuffix(",")
                if not cpu:
                    eval_group["CUDA_VISIBLE_DEVICES"] = gpu_indices
                num_processes = len(gpu_indices.split(","))
                if args.O1:
                    eval_group["OMP_NUM_THREADS"] = str(os.cpu_count() // num_processes)

                _,  _accelerate_config_file = modify_config_file(accelerate_config_file, num_processes, port=port+1, copy=True)
                async_cmd = f"accelerate launch --config_file={_accelerate_config_file} {file} {extra_args}"
                
                async_cmd = get_cmd_as_list(async_cmd)
                process2 = subprocess.Popen(async_cmd, env=eval_group, start_new_session=True)

                process2.wait()
                process1.wait()
            else:
                process1.wait()
        except KeyboardInterrupt:
            print("\nTerminating subprocesses...")
            # Send SIGTERM to entire process groups
            processes = [process1]
            if process2 is not None:
                processes.append(process2)
            for p in processes:
                if p and p.poll() is None:
                    os.killpg(os.getpgid(p.pid), signal.SIGTERM)
            sys.exit(1)

